/*
 * GammaCorrectTest.cpp - Version 0.3 - March 30, 2020
 *
 * Exploratory program to try out gamma correction
 * Allows dynamically adjusting gamma correction
 *   with any of Red, Green, Blue, Yellow, Cyan, Magenta and White.
 *
 * Author: Sam Buss
 *
 * Software accompanying POSSIBLE SECOND EDITION TO the book
 *		3D Computer Graphics: A Mathematical Introduction with OpenGL,
 *		by S. Buss, Cambridge University Press, 2003.
 *
 * Software is "as-is" and carries no warranty.  It may be used without
 *   restriction, but if you modify it, please change the filenames to
 *   prevent confusion between different versions.
 * Bug reports: Sam Buss, sbuss@ucsd.edu.
 * Web page: http://math.ucsd.edu/~sbuss/MathCG2
 */

// See printf statements in main program for keyboard controls

// These libraries are needed to link the program (Visual Studio specific)
#pragma comment(lib,"opengl32.lib")
#pragma comment(lib,"glu32.lib")
#pragma comment(lib,"glfw3.lib")
#pragma comment(lib,"glew32s.lib")
#pragma comment(lib,"glew32.lib")

// Use the static library (so glew32.dll is not needed):
#define GLEW_STATIC
#include <GL/glew.h> 
#include <GLFW/glfw3.h>

#include "GlShaderMgr.h"
bool check_for_opengl_errors();     // Function prototype (should really go in a header file)

// Enable standard input and output via printf(), etc.
// Put this include *after* the includes for glew and GLFW!
#include <stdio.h>

// ************************
// General data helping with setting up VAO (Vertex Array Objects)
//    and Vertex Buffer Objects.
// ***********************

const int NumObjects = 1;
const int iRect = 0;

unsigned int myVBO[NumObjects];  // Vertex Buffer Object - holds an array of data
unsigned int myVAO[NumObjects];  // Vertex Array Object - holds info about how the vertex data is formatted

// State info

bool sRGBmode = false;
bool useBordersMode = true;

float gamma = 1.0;                    // Current gamma value
float curColor[3] = { 1.0, 1.0, 1.0 };
int colorCode = 7;
char colorNames[8][16] = { "Blue", "Green", "Cyan", "Red", "Magenta", "Yellow", "White", "Ignored by sRGB" };

float dragMode = false;
float startX;
float startGamma;
int windowWidth;

// We create one shader program: it consists of a vertex shader and a fragment shader
unsigned int shaderProgram1;
const unsigned int vertPos_loc = 0;   // Corresponds to "location = 0" in the verter shader definition
int screenDims_loc;                   // Location of screenDims (uniform vec2) in shader program
int gamma_loc;                        // Location of gamma (uniform float) in shader program
int color_loc;                        // Location of saturated RGB color (uniform vec3) in shader program
int sRGB_loc;                         // Location of sRGB (uniform bool) in shader program
int useBorders_loc;                   // Location of useBorders (uniform bool) in shader program   

// *************************
// mySetupGeometries defines the scene data, especially vertex  positions and colors.
//    - It also loads all the data into the VAO's (Vertex Array Objects) and
//      into the VBO's (Vertex Buffer Objects).
// This routine is only called once to initialize the data.
// *************************
void mySetupGeometries() {
	// Set up VAO and VBO for a single rectangle [-1,1]x[-1,1] drawn as two triangles

    // Allocate Vertex Array Objects (VAOs) and Vertex Buffer Objects (VBOs).
    glGenVertexArrays(NumObjects, &myVAO[0]);
    glGenBuffers(NumObjects, &myVBO[0]);

    // FIRST GEOMETRY: One rectangle vertices ordered for triangle strip
	float rectVerts[] = {
		-1.0f, -1.0f,    1.0f, -1.0f,  -1.0f, 1.0f,   1.0f, 1.0f 
	};

    // Bind (and initialize) the Vertex Array Object
    glBindVertexArray(myVAO[iRect]);

    // Allocate space in the vertex buffer (the VBO)
    //   and load the four pairs of x,y values into the VBO.  
    glBindBuffer(GL_ARRAY_BUFFER, myVBO[iRect]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(rectVerts), rectVerts, GL_STATIC_DRAW);

    // Bind the Vertex Array Object and then configure vertex attributes(s).
    glVertexAttribPointer(vertPos_loc, 2, GL_FLOAT, GL_FALSE, 0, (void*)0);	
    glEnableVertexAttribArray(vertPos_loc);

    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}
// *************************************
// Main routine for rendering the scene
// myRenderScene() is called every time the scene needs to be redrawn.
// mySetupGeometries() has already created the vertex and buffer objects, etc.
// setup_shaders() has already created the shader programs.
// gamma, color, and screenDims values are already set in the shader program.
// *************************************
void myRenderScene() {

	// Clear the rendering window
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		
    // Draw rectangle covering whole window, as two triangles
    // Colors for the triangles have already been loaded into the VBO.
	glBindVertexArray(myVAO[iRect]);
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    glBindVertexArray(0);       // Not necessary, but a good idea
    check_for_opengl_errors();  // Really a great idea to check for errors -- esp. good for debugging!
}

void my_setup_SceneData() {
	mySetupGeometries();
	
    GlShaderMgr::LoadShaderSource("GammaCorrectTest.glsl");
    shaderProgram1 = GlShaderMgr::CompileAndLinkAll();
    color_loc = glGetUniformLocation(shaderProgram1, "fullColor");
    gamma_loc = glGetUniformLocation(shaderProgram1, "gamma");
    screenDims_loc = glGetUniformLocation(shaderProgram1, "screenDims");
    sRGB_loc = glGetUniformLocation(shaderProgram1, "sRGB");
    useBorders_loc = glGetUniformLocation(shaderProgram1, "useBorders");

    glUseProgram(shaderProgram1);
    glUniform1f(gamma_loc, gamma);
    glUniform3fv(color_loc, 1, curColor);
    glUniform1i(sRGB_loc, sRGBmode);
    glUniform1i(useBorders_loc, useBordersMode);
    
    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}

// *******************************************************
// Process all key press events.
// This routine is called each time a key is pressed or released or repeats.
// *******************************************************
void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods) {
	if (action == GLFW_RELEASE) {
		return;			// Ignore key up (key release) events
	}
	if (key == GLFW_KEY_ESCAPE ) {
		glfwSetWindowShouldClose(window, true);
        return;
	}
    if (key == GLFW_KEY_S) {
        sRGBmode = !sRGBmode;
        printf("sRGB mode turned %s.\n", sRGBmode ? "on" : "off");
        glUniform1i(sRGB_loc, sRGBmode);
        return;
    }
    if (key == GLFW_KEY_U) {
        useBordersMode = !useBordersMode;
        glUniform1i(useBorders_loc, useBordersMode);
        return;
    }
    if (key == GLFW_KEY_L) {        // Enter linear mode
        glUniform1i(sRGB_loc, sRGBmode);
        sRGBmode = false;
        gamma = 1;
        glUniform1i(sRGB_loc, sRGBmode);
        glUniform1f(gamma_loc, gamma);
        printf("Linear mode (gamma = 1 and sRGB is not used).\n");
        return;
    }
    if (key == GLFW_KEY_R) {
        colorCode = 4;
    }
    else if (key == GLFW_KEY_G) {
        colorCode = 2;
    }
    else if (key == GLFW_KEY_B) {
        colorCode = 1;
    }
    else if (key == GLFW_KEY_C) {
        colorCode = 3;
    }
    else if (key == GLFW_KEY_M) {
        colorCode = 5;
    }
    else if (key == GLFW_KEY_Y) {
        colorCode = 6;
    }
    else if (key == GLFW_KEY_W) {
        colorCode = 7;
    }
    else {
        return;
    }
    curColor[0] = (colorCode & 0x04) ? 1.0f : 0.0f;
    curColor[1] = (colorCode & 0x02) ? 1.0f : 0.0f;
    curColor[2] = (colorCode & 0x01) ? 1.0f : 0.0f;
    glUniform3f(color_loc, curColor[0], curColor[1], curColor[2]);
    check_for_opengl_errors();
}

// *******************************************************
// Process all mouse button events.
// This routine is called each time a mouse button is pressed or released.
// *******************************************************
void mouse_button_callback(GLFWwindow* window, int button, int action, int mods)
{
    if (button != GLFW_MOUSE_BUTTON_LEFT) {
        return;
    }
    if (action == GLFW_PRESS) {
        double xpos, ypos;
        glfwGetCursorPos(window, &xpos, &ypos);
        // Scale x and y values into the range [-1,1]. 
        // Note that y values are negated since mouse measures y position from top of the window
        startX = (2.0f*(float)xpos / (float)(windowWidth - 1)) - 1.0f;
        startGamma = gamma;
        dragMode = true;
    }
    else if (action == GLFW_RELEASE) {
        dragMode = false;
        printf("Gamma = %7.5f (%s).\n", gamma, sRGBmode ? "Ignored by sRGB" : &colorNames[colorCode-1][0]);
    }
}

void cursor_pos_callback(GLFWwindow* window, double x, double y) {
    if ( !dragMode ) {
        return;
    }
    float curX = (2.0f*(float)x / (float)(windowWidth - 1)) - 1.0f;
    float d1 = startX + 1.0f;
    float d2 = 1.0f - startX;
    float denom = d2 > d1 ? d2 : d1;
    gamma = startGamma * (1.0f + (curX - startX) / denom);
    gamma = (gamma < 0.1f) ? 0.1f : gamma;
    gamma = (gamma >3.0f) ? 3.0f : gamma;
    glUniform1f(gamma_loc, gamma);
    return;
}


// *************************************************
// This function is called when the graphics window is first created,
//    and again whenever it is resized.
// The Projection View Matrix is typically set here.
//    But this program does not use any transformations or matrices.
// *************************************************
void window_size_callback(GLFWwindow* window, int width, int height) {
	glViewport(0, 0, width, height);		// Draw into entire window
    windowWidth = width;

    if (glIsProgram(shaderProgram1)) {
        glUniform2f(screenDims_loc, (float)width, (float)height);
    }
    check_for_opengl_errors();
}

void my_setup_OpenGL() {
	
	glEnable(GL_DEPTH_TEST);	// Enable depth buffering
	glDepthFunc(GL_LEQUAL);		// Useful for multipass shaders

	// Set polygon drawing mode for front and back of each triangle
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	
}

void error_callback(int error, const char* description)
{
	// Print error
	fputs(description, stderr);
}

void setup_callbacks(GLFWwindow* window) {
	// Set callback function for resizing the window
	glfwSetFramebufferSizeCallback(window, window_size_callback);

	// Set callback for key up/down/repeat events
	glfwSetKeyCallback(window, key_callback);

	// Set callbacks for mouse movement (cursor position) and mouse botton up/down events.
	glfwSetCursorPosCallback(window, cursor_pos_callback);
	glfwSetMouseButtonCallback(window, mouse_button_callback);
}

// **********************
// Here is the main program
// **********************

int main() {
	glfwSetErrorCallback(error_callback);	// Supposed to be called in event of errors. (doesn't work?)
    glfwInit();

#if defined(__APPLE__) || defined(__linux__)
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
	glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
#endif

	const int initWidth = 800;
	const int initHeight = 600;
	GLFWwindow* window = glfwCreateWindow(initWidth, initHeight, "GammaCorrectTest (& sRGB)", NULL, NULL);
	if (window == NULL) {
		fprintf(stdout, "Failed to create GLFW window!\n");
		glfwTerminate();
		return -1;
	}
	glfwMakeContextCurrent(window);

	glewInit();

	// Print info of GPU and supported OpenGL version
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("OpenGL version supported %s\n", glGetString(GL_VERSION));
#ifdef GL_SHADING_LANGUAGE_VERSION
	printf("Supported GLSL version is %s.\n", (char *)glGetString(GL_SHADING_LANGUAGE_VERSION));
#endif
    printf("Using GLEW version %s.\n", glewGetString(GLEW_VERSION));
	printf("------------------------------\n");
    printf("Press mouse button, drag left & right to adjust gamma correction value.\n");
    printf("new gamma value is printed in text window when mouse button is released.\n");
    printf("Press 'S' (sRGB) to toggle using sRGB (gamma is ignored).\n");
    printf("Press 'L' (linear) to set gamma=1 (turns off sRGB).\n");
    printf("Press 'R', 'G', 'B', 'C', 'Y', 'M', 'W' to choose color.\n");
    printf("Press 'U', to toggle use of borders.\n");
    printf("Press Escape to exit.\n");
	
    setup_callbacks(window);

	// Initialize OpenGL, the scene and the shaders
    my_setup_OpenGL();

	my_setup_SceneData();
 
     // Call window_size_callback after shader is set up, so as 
    //   to give the shader program the screen dimensions
    window_size_callback(window, initWidth, initHeight);
    check_for_opengl_errors();

    // Loop while program is not terminated.
	while (!glfwWindowShouldClose(window)) {
	
		myRenderScene();				// Render into the current buffer
		glfwSwapBuffers(window);		// Displays what was just rendered (using double buffering).

		// Poll events (key presses, mouse events)
		glfwWaitEvents();					// Use this if no animation.
		// glfwWaitEventsTimeout(1.0/60.0);	// Use this to animate at 60 frames/sec (timing is NOT reliable)
		// glfwPollEvents();				// Use this version when animating as fast as possible
	}

	glfwTerminate();
	return 0;
}

// If an error is found, it could have been caused by any command since the
//   previous call to check_for_opengl_errors()
// To find what generated the error, you can try adding more calls to
//   check_for_opengl_errors().
char errNames[9][36] = {
    "Unknown OpenGL error",
    "GL_INVALID_ENUM", "GL_INVALID_VALUE", "GL_INVALID_OPERATION",
    "GL_INVALID_FRAMEBUFFER_OPERATION", "GL_OUT_OF_MEMORY",
    "GL_STACK_UNDERFLOW", "GL_STACK_OVERFLOW", "GL_CONTEXT_LOST" };
bool check_for_opengl_errors() {
    int numErrors = 0;
    GLenum err;
    while ((err = glGetError()) != GL_NO_ERROR) {
        numErrors++;
        int errNum = 0;
        switch (err) {
        case GL_INVALID_ENUM:
            errNum = 1;
            break;
        case GL_INVALID_VALUE:
            errNum = 2;
            break;
        case GL_INVALID_OPERATION:
            errNum = 3;
            break;
        case GL_INVALID_FRAMEBUFFER_OPERATION:
            errNum = 4;
            break;
        case GL_OUT_OF_MEMORY:
            errNum = 5;
            break;
        case GL_STACK_UNDERFLOW:
            errNum = 6;
            break;
        case GL_STACK_OVERFLOW:
            errNum = 7;
            break;
        case GL_CONTEXT_LOST:
            errNum = 8;
            break;
        }
        printf("OpenGL ERROR: %s.\n", errNames[errNum]);
    }
    return (numErrors != 0);
}