// ***************************
// Shader programs to be used for gamma correction
//     experimentation.
// Basic functionality: Colors are either
//     (a) (alpha C)^\gamma, or
//     (b) Interspersed with 
           fraction alpha of pixels set to color C,
//         and fraction (1-alpha) set to 0 (Black)
// Color C intended to be any fully saturated color, i.e.,
//     Red, Green, Blue, Cyan, Magenta, Yellow, or White
//     (1,0,0), (0,1,0), (0,0,1), (0,1,1), (1,0,1), (1,1,0), (1,1,1).
// Right half of window uses option (a)
// Left half of window uses option (b).
// Value of alpha will be 0, 0.25, 0.5, 0.75, or 1.0
//     in five horizontal bands across the window.
//
// Revision: March 31, 2020
// ***************************

// This is the "no-op" vertex shader
#beginglsl vertexShader gct_VertexShader
#version 330 core
layout (location = 0) in vec3 vertPos;         // Position in attribute location 0

void main()
{
    gl_Position = vec4(vertPos.x, vertPos.y, vertPos.z, 1.0);
}
#endglsl

#beginglsl fragmentshader gammaTester_frag
#version 330 core

uniform float gamma;            // gamma (exponent)
uniform vec3 fullColor;         // Full intensity color
uniform vec2 screenDims;        // Width and height of the screen
uniform bool useBorders = true; // True if should separate with black borders

uniform bool sRGB = true;		// True if should use sRGB instead of directly gamma.

out vec4 fragColor;             // Final pixel color

bool turnon( int ixPixel, int iyPixel, int onLevel );
bool inBorder( float xPixel, float yPixel );
bool closeTo( float a, float b);
float linear_to_sRGB ( float linVal );

void main() {
    float xPixel = gl_FragCoord.x;
    float yPixel = gl_FragCoord.y;
    // onLevel: intensity value is 0, 1, 2, 3, or 4.
    int onLevel = int(gl_FragCoord.y*5.0/(screenDims.y+1.0));
    float alpha = onLevel*0.25;
	if ( useBorders && inBorder(xPixel, yPixel) ) {
	    fragColor = vec4(0.0, 0.0, 0.0, 1.0);
		return;
	}
    if ( xPixel > screenDims.x/2.0 ) {
       // On righthand side, use gamma corrected color or sRGB color.
	   if ( !sRGB ) {
            fragColor = vec4(fullColor*pow(alpha,gamma), 1.0); // Gamma correction
		}
		else {
		    fragColor = vec4(fullColor*linear_to_sRGB(alpha),1.0);	// Linear to sRGB conversion
		}
        return;
    }
	if ( turnon( int(xPixel), int(yPixel), onLevel ) ) {
        fragColor = vec4(fullColor, 1.0);
    }
    else {
        fragColor = vec4(0.0, 0.0, 0.0, 1.0);
    } 
}

bool turnon( int ixPixel, int iyPixel, int onLevel ) {
    if ( onLevel==0 ) {
        return false;
    }
    if ( onLevel==4 ) {
        return true;
    }
	int ixPixel2 = ixPixel%2;
	if ( onLevel==2 ) {
	    return ixPixel2 == ((iyPixel>>1) + iyPixel)%2;
	}
	int inCorner = 3;
	if ( ((iyPixel%2) == 0) && (ixPixel2 == ((iyPixel>>1)%2)) ) {
	    inCorner = 1;
	}
	return inCorner==onLevel;
}

bool inBorder( float xPixel, float yPixel ) {
    return ( closeTo(xPixel, screenDims.x/2.0)
	     || closeTo(yPixel, screenDims.y*0.4)
		 || closeTo(yPixel, screenDims.y*0.6)
		 || closeTo(yPixel, screenDims.y*0.8)
		 || closeTo(yPixel, screenDims.y) );
}

bool closeTo( float a, float b) {
    float diff = a-b;
	return (diff<2.0 && diff>-2.0);
}

float linear_to_sRGB( float linVal ) {
    if ( linVal<0.0031308 ) {
	    return 12.92*linVal;
	}
	return 1.055*pow(linVal,1/2.4)-0.055;
}
#endglsl

